
rem The file to compile is MAIN.BAS

#define neant     0
#define pion      1
#define cavalier  2
#define fou       3
#define tour      4
#define dame      5
#define roi       6
#define noir     -1
#define blanc     1
#define e8c8      0
#define e8g8      1
#define e1c1      2
#define e1g1      3

enum boolean
  vrai = -1
  faux
end enum

type coup
  x1 as byte
  y1 as byte
  x2 as byte
  y2 as byte
end type

type position
  damier(7, 7) as byte
  trait        as byte
  roque(3)     as byte
  passant      as byte
  demiCoups    as byte
  pli          as short
  valeur       as short
end type

dim shared courante as position
dim shared initiale(7, 7) as byte
dim shared sGame as string

declare function coupToStr(c as coup) as string
declare sub strToCoup(s as string, byref c as coup)
declare sub initialiseJeu
declare function passage(c as const coup, p as const position) as byte
declare function geometrie(c as const coup, p as const position) as byte
declare function echec(p as const position) as boolean
declare function empechement(p as position, r as byte) as byte
declare sub mouvement(c as const coup, byref p as position)
declare function possibles(p as position) as string
declare sub evaluation(byref p as position)
declare function meilleurCoup as coup
declare function etatInitial(x as byte, y as byte) as boolean
declare function existe(x as integer, y as integer) as boolean
declare function pionVoisin(c as coup) as boolean
declare function admissibles(liste as string) as string
declare function meilleurs(liste as string) as string

function coupToStr(c as coup) as string
  return chr(c.x1 + 97) + chr(c.y1 + 49) + chr(c.x2 + 97) + chr(c.y2 + 49)
end function

sub strToCoup(s as string, byref c as coup)
  c.x1 = s[0] - 97
  c.y1 = s[1] - 49
  c.x2 = s[2] - 97
  c.y2 = s[3] - 49
end sub

sub initialiseJeu
  '
  restore
  '
  for y as byte = 7 to 0 step -1
    for x as byte = 0 to 7
      read initiale(x, y)
      courante.damier(x, y) = initiale(x, y)
    next x
  next y
  '
  courante.trait = blanc
  '
  courante.roque(e8c8) = vrai
  courante.roque(e8g8) = vrai
  courante.roque(e1c1) = vrai
  courante.roque(e1g1) = vrai
  '
  courante.passant = 8
  ' valeurs de 0  7 : colonne
  ' valeur 8 : aucun passant
  '
  courante.demiCoups = 0
  courante.pli = 1
  '
  sGame = ""
  '
end sub

data -4,-2,-3,-5,-6,-3,-2,-4
data -1,-1,-1,-1,-1,-1,-1,-1
data  0, 0, 0, 0, 0, 0, 0, 0
data  0, 0, 0, 0, 0, 0, 0, 0
data  0, 0, 0, 0, 0, 0, 0, 0
data  0, 0, 0, 0, 0, 0, 0, 0
data  1, 1, 1, 1, 1, 1, 1, 1
data  4, 2, 3, 5, 6, 3, 2, 4

function passage(c as const coup, p as const position) as byte
  '
  dim x as byte = c.x1
  dim y as byte = c.y1
  dim vx as byte = sgn(c.x2 - c.x1)
  dim vy as byte = sgn(c.y2 - c.y1)
  dim obstacle as boolean = faux
  dim arrivee as boolean = faux
  '
  do
    x += vx
    y += vy
    if not(p.damier(x, y) = neant) then obstacle = vrai
    if (x = c.x2) and (y = c.y2) then arrivee = vrai
  loop until (obstacle or arrivee)
  '
  return arrivee
  '
end function

function geometrie(c as const coup, p as const position) as byte
  '
  if (c.x1 = c.x2) andalso (c.y1 = c.y2) then return faux
  '
  if sgn(p.damier(c.x2, c.y2)) = sgn(p.damier(c.x1, c.y1)) then return faux
  '
  select case as const abs(p.damier(c.x1, c.y1))
  '
  case pion
    if sgn(c.y2 - c.y1) = p.damier(c.x1, c.y1) then
      '
      if (abs(c.x2 - c.x1) = 1) andalso (c.y2 - c.y1 = p.damier(c.x1, c.y1)) then
        if p.damier(c.x2, c.y2) = neant then
          if p.passant = 8 then
            return faux
          else
            if c.x2 = p.passant then
              if (p.damier(c.x1, c.y1) = blanc) and (c.y1 = 4) then return vrai
              if (p.damier(c.x1, c.y1) = noir) and (c.y1 = 3) then return vrai
            end if
          end if
        else
          return vrai
        end if
      end if
      '
      if (c.x2 = c.x1) andalso (p.damier(c.x2, c.y2) = neant) then
        if (c.y2 - c.y1 = p.damier(c.x1, c.y1)) then return vrai
        if (c.y2 - c.y1 = 2 * p.damier(c.x1, c.y1)) then
          if p.damier(c.x1, c.y1 + p.damier(c.x1, c.y1)) = neant then
            if (p.damier(c.x1, c.y1) = blanc) andalso (c.y1 = 1) then return vrai
            if (p.damier(c.x1, c.y1) = noir) andalso (c.y1 = 6) then return vrai
          end if
        end if
      end if
      '
    end if
    '
  case cavalier
    if (c.x2 - c.x1) * (c.x2 - c.x1) + (c.y2 - c.y1) * (c.y2 - c.y1) = 5 then return vrai
    '
  case fou
    if (c.x2 - c.x1) * (c.x2 - c.x1) = (c.y2 - c.y1) * (c.y2 - c.y1) then
      if passage(c, p) then return vrai
    end if
    '
  case tour
    if (c.x2 - c.x1) * (c.x2 - c.x1) * (c.y2 - c.y1) * (c.y2 - c.y1) = 0 then
      if passage(c, p) then return vrai
    end if
    '
  case dame
    if (c.x2 - c.x1) * (c.x2 - c.x1) = (c.y2 - c.y1) * (c.y2 - c.y1) then
      if passage(c, p) then return vrai
    end if
    '
    if (c.x2 - c.x1) * (c.x2 - c.x1) * (c.y2 - c.y1) * (c.y2-c.y1) = 0 then
      if passage(c, p) then return vrai
    end if
    '
  case roi
    if (c.x2 - c.x1) * (c.x2 - c.x1) + (c.y2 - c.y1) * (c.y2 - c.y1) <= 2 then return vrai
    '
    if (c.y2 = c.y1) and (abs(c.x2 - c.x1) = 2) then
      '
      if c.x2 = 2 then
        if p.damier(1, c.y1) = neant then
          if p.damier(2, c.y1) = neant then
            if p.damier(3, c.y1) = neant then
              if (p.damier(c.x1, c.y1) = roi * blanc) and p.roque(e1c1) then return vrai
              if (p.damier(c.x1, c.y1) = roi * noir) and p.roque(e8c8) then return vrai
            end if
          end if
        end if
      end if
      '
      if c.x2 = 6 then
        if p.damier(5, c.y1) = neant then
          if p.damier(6, c.y1) = neant then
            if (p.damier(c.x1, c.y1) = roi * blanc) and p.roque(e1g1) then return vrai
            if (p.damier(c.x1, c.y1) = roi * noir) and p.roque(e8g8) then return vrai
          end if
        end if
      end if
      '
    end if
    '
  end select
  '
  return faux
  '
end function

function echec(p as const position) as boolean
  '
  dim as byte x1, y1, x2, y2
  y2 = 7
  x2 = 0
  while not(p.damier(x2, y2) = roi * p.trait)
    x2 += 1
    if x2 = 8 then
      x2 = 0
      y2 -= 1
    end if
  wend
  '
  dim c as coup
  '
  c.x2 = x2
  c.y2 = y2
  for x1 = 0 to 7
    for y1 = 0 to 7
      if sgn(p.damier(x1, y1)) = -1 * p.trait then
        c.x1 = x1
        c.y1 = y1
        if geometrie(c, p) then return vrai
      end if
    next y1
  next x1
  '
  return faux
  '
end function

function empechement(p as position, r as byte) as byte
  '
  dim as byte x1, y1, x2, y2
  dim as byte a, b
  '
  select case as const r
  case e8c8
    y2 = 7
    a = 0
    b = 4
  case e8g8
    y2 = 7
    a = 4
    b = 7
  case e1c1
    y2 = 0
    a = 0
    b = 4
  case e1g1
    y2 = 0
    a = 4
    b = 7
  end select
  '
  dim c as coup
  '
  c.y2 = y2
  for x2 = a to b
    c.x2 = x2
    for x1 = 0 to 7
      for y1 = 0 to 7
        if sgn(p.damier(x1, y1)) = -1 * p.trait then
          c.x1 = x1
          c.y1 = y1
          if geometrie(c, p) then return vrai
        end if
      next y1
    next x1
  next x2
  '
  return faux
  '
end function

sub mouvement(c as const coup, byref p as position)
  '
  if (abs(p.damier(c.x1, c.y1)) = pion) and ((c.y2 = 7) or (c.y2 = 0)) then
    p.damier(c.x1, c.y1) *= dame
  end if
  '
  if (abs(p.damier(c.x1, c.y1)) = roi) and (abs(c.x2 - c.x1) = 2) then
    if c.x2 = 2 then
      p.damier(3, c.y1) = p.damier(0, c.y1)
      p.damier(0, c.y1) = neant
    else
      p.damier(5, c.y1) = p.damier(7, c.y1)
      p.damier(7, c.y1) = neant
    end if
  end if
  '
  if (abs(p.damier(c.x1, c.y1)) = pion) andalso (abs(c.x2 - c.x1) = 1) then
    if p.damier(c.x2, c.y2) = neant then
      p.damier(c.x2, c.y1) = neant
    end if
  end if
  '
  if (abs(p.damier(c.x1, c.y1)) = roi) then
    if p.trait = blanc then
      p.roque(e1c1) = faux
      p.roque(e1g1) = faux
    else
      p.roque(e8c8) = faux
      p.roque(e8g8) = faux
    end if
  end if
  '
  if (abs(p.damier(c.x1, c.y1)) = tour) then
    if (c.x1 = 0) andalso (c.y1 = 7) then p.roque(e8c8) = faux
    if (c.x1 = 7) andalso (c.y1 = 7) then p.roque(e8g8) = faux
    if (c.x1 = 0) andalso (c.y1 = 0) then p.roque(e1c1) = faux
    if (c.x1 = 7) andalso (c.y1 = 0) then p.roque(e1g1) = faux
  end if
  '
  if (abs(p.damier(c.x2, c.y2)) = tour) then
    if (c.x2 = 0) andalso (c.y2 = 7) then p.roque(e8c8) = faux
    if (c.x2 = 7) andalso (c.y2 = 7) then p.roque(e8g8) = faux
    if (c.x2 = 0) andalso (c.y2 = 0) then p.roque(e1c1) = faux
    if (c.x2 = 7) andalso (c.y2 = 0) then p.roque(e1g1) = faux
  end if
  '
  if (abs(p.damier(c.x1, c.y1)) = pion) and (abs(c.y2 - c.y1) = 2) then
    p.passant = c.x1
  else
    p.passant = 8
  end if
  '
  if (abs(p.damier(c.x1, c.y1)) = pion) or (p.damier(c.x2, c.y2) <> neant) then
    p.demiCoups = 0
  else
    p.demiCoups += 1
  end if
  '
  if p.trait = noir then p.pli += 1
  '
  p.damier(c.x2, c.y2) = p.damier(c.x1, c.y1)
  p.damier(c.x1, c.y1) = neant
  '
  p.trait *= -1
  '
end sub

function possibles(p as position) as string
  '
  dim s as string = ""
  dim as byte x1, y1, x2, y2
  dim c as coup
  dim cs as string
  dim fictive as position
  '
  for x1 = 0 to 7
    c.x1 = x1
    for y1 = 0 to 7
      c.y1 = y1
      if sgn(p.damier(x1, y1)) = p.trait then
        for x2 = 0 to 7
          c.x2 = x2
          for y2 = 0 to 7
            c.y2 = y2
            if geometrie(c, p) then
              cs = coupToStr(c)
              fictive = p
              mouvement(c, fictive)
              fictive.trait *= -1
              if echec(fictive) then cs = ""
              if abs(p.damier(x1, y1)) = roi then
                if abs(c.x2 - c.x1) = 2 then
                  select case cs
                  case "e8c8"
                    if not(empechement(p, e8c8)) then
                      s += cs
                    end if
                  case "e8g8"
                    if not(empechement(p, e8g8)) then
                      s += cs
                    end if
                  case "e1c1"
                    if not(empechement(p, e1c1)) then
                      s += cs
                    end if
                  case "e1g1"
                    if not(empechement(p, e1g1)) then
                      s += cs
                    end if
                  end select
                else
                  s += cs
                end if
              else
                s += cs
              end if
            end if
          next y2
        next x2
      end if
    next y1
  next x1
  '
  possibles = s
  '
end function

dim shared prix(roi * noir to roi * blanc) as byte

prix(roi * noir)       = -99
prix(dame * noir)      = -12
prix(tour * noir)      = -07
prix(fou * noir)       = -04
prix(cavalier * noir)  = -04
prix(pion * noir)      = -01
prix(neant)            =  00
prix(pion * blanc)     =  01
prix(cavalier * blanc) =  04
prix(fou * blanc)      =  04
prix(tour * blanc)     =  07
prix(dame * blanc)     =  12
prix(roi * blanc)      =  99

sub evaluation(byref p as position)
  '
  dim as byte x, y
  '
  p.valeur = 0
  '
  for x = 0 to 7
    for y = 0 to 7
      p.valeur += prix(p.damier(x, y)) * p.trait
    next y
  next x
  '
end sub

function etatInitial(byval x as byte, byval y as byte) as boolean
  '
  if courante.damier(x, y) = initiale(x, y) then
    return vrai
  else
    return faux
  end if
  '
end function

function existe(x as integer, y as integer) as boolean
  '
  if (x > -1) and (x < 8) and (y > -1) and (y < 8) then
    return vrai
  else
    return faux
  end if
  '
end function

function pionVoisin(c as coup) as boolean
  '
  dim as byte v(7, 1) = {{-1, 0}, {-1,-1}, {0,-1}, {1,-1}, {1, 0}, {1, 1}, {0, 1}, {-1, 1}}
  dim as integer x, y
  dim temp as boolean = faux
  '
  for i as integer = 0 to 7
    x = c.x2 + v(i, 0)
    y = c.y2 + v(i, 1)
    if existe(x, y) andalso courante.damier(x, y) = courante.damier(c.x1, c.y1) then
      if (x <> c.x1) or (y <> c.y1) then temp = vrai
    end if
  next i
  '
  return temp
end function

function admissibles(liste as string) as string
  '
  if liste = "" then return liste
  '
  dim i as integer
  dim cs as string
  dim c as coup
  dim temp as string = ""
  dim as integer note(1 to len(liste)\4)
  '
  for i = 1 to len(liste)\4
    '
    cs = mid(liste, 1 + 4 * (i - 1), 4)
    strToCoup(cs, c)
    note(i) = 0
    '
    select case as const abs(courante.damier(c.x1, c.y1))
    case pion
      if (c.x1 mod 7 = 0) or _
      (not pionVoisin(c)) then note(i) = -1
    case cavalier
      if (c.x2 mod 7 = 0) and _
      (courante.damier(c.x2, c.y2) = neant) then note(i) = -1
    case fou
      rem
    case tour
      if courante.pli <= 8 then note(i) = -1
    case dame
      if courante.pli <= 8 then note(i) = -1
    case roi
      if (courante.damier(c.x2, c.y2) = neant)_
      and (not (abs(c.x2 - c.x1) = 2))_
      and (not echec(courante))_
      then note(i) = -1
    end select
    '
    if initiale(c.x2, c.y2) = courante.damier(c.x1, c.y1) then note(i) = -1
    '
  next i
  '
  for i = 1 to len(liste)\4
    cs = mid(liste, 1 + 4 * (i - 1), 4)
    if note(i) = 0 then temp = temp + cs
  next i
  '
  return temp
  '
end function

function meilleurs(liste as string) as string
  '
  dim as string pss1, pss2, pss3
  dim as coup c0, c1, c2, c3
  dim as string*4 cs0, cs1, cs2, cs3
  dim as position fictive, f1, f2, f3
  dim as byte i, j, k, l
  dim as string temp = ""
  dim as integer note(1 to len(liste)\4)
  dim as integer noteMax = -100
  dim as integer d0, d1, d2, d3
  '
  dim as boolean d0_plus_d1_toujours_positif
  dim as boolean d0_plus_d1_plus_d2_toujours_negatif
  '
  evaluation(courante)
  '
  for i = 1 to len(liste)\4
    '
    note(i) = 0
    '
    cs0 = mid(liste, 4 * i - 3, 4)
    strToCoup(cs0, c0)
    fictive = courante
    mouvement(c0, fictive)
    evaluation(fictive)
    '
    d0 = noir * fictive.valeur - courante.valeur
    rem sgn(d0) = 0 ou +1
    '
    d0_plus_d1_toujours_positif = vrai
    '
    pss1 = possibles(fictive)
    for j = 1 to len(pss1)\4
      '
      cs1 = mid(pss1, 4 * j - 3, 4)
      strToCoup(cs1, c1)
      f1 = fictive
      mouvement(c1, f1)
      evaluation(f1)
      '
      d1 = f1.valeur - noir * fictive.valeur
      rem sgn(d1) = 0 ou -1
      if d0 + d1 <= 0 then d0_plus_d1_toujours_positif = faux
      '
      d0_plus_d1_plus_d2_toujours_negatif = vrai
      '
      pss2 = possibles(f1)
      for k = 1 to len(pss2)\4
        cs2 = mid(pss2, 4 * k - 3, 4)
        strToCoup(cs2, c2)
        f2 = f1
        mouvement(c2, f2)
        evaluation(f2)
        '
        d2 = noir * f2.valeur - f1.valeur
        rem sgn(d2) = 0 ou +1
        if d0 + d1 + d2 >= 0 then d0_plus_d1_plus_d2_toujours_negatif = faux
        '
        pss3 = possibles(f2)
        for l = 1 to len(pss3)\4
          cs3 = mid(pss3, 4 * l - 3, 4)
          strToCoup(cs3, c3)
          f3 = f2
          mouvement(c3,f3)
          evaluation(f3)
          '
          d3 = f3.valeur - noir * f2.valeur
          rem sgn(d3) = 0 ou -1
          '
        next l
      next k
      '
      if d0_plus_d1_plus_d2_toujours_negatif = vrai then
        '
        note(i) -= 1
        '
      end if
      '
    next j
    '
    if d0_plus_d1_toujours_positif then
      '
      note(i) += d0 * (100 - abs(prix(courante.damier(c0.x1, c0.y1))))
      '
    end if
    '
    #IFDEF DEV
    shell("echo." & cs0 & " " & str(note(i)))
    #ENDIF
    '
  next i
  '
  for i = 1 to len(liste)\4
    if note(i) > noteMax then noteMax = note(i)
  next i
  '
  for i = 1 to len(liste)\4
    cs0 = mid(liste, 4 * i - 3, 4)
    if note(i) = noteMax then temp = temp + cs0
  next i
  '
  return temp
  '
end function

function meilleurCoup as coup
  '
  dim as string liste1 = possibles(courante)
  dim as string liste2 = admissibles(liste1)
  '
  if liste2 = "" and liste1 <> "" then liste2 = liste1
  '
  dim as string liste3 = meilleurs(liste2)
  '
  dim temp as coup
  dim cBook as coup
  '
  temp.x1 = 0
  temp.y1 = 0
  temp.x2 = 0
  temp.y2 = 0
  '
  if liste3 <> "" then strToCoup(mid(liste3, 1, 4), temp)
  '
  dim as uByte f = freeFile
  dim as integer erreur = open("Book.txt" for input as #f)
  dim as string l
  '
  if erreur = 0 then
    do until eof(f)
      line input #f, l
      if (instr(1, l, sGame) = 1) andalso len(l) > len(sGame) then
        strToCoup(mid(l, len(sGame)+1, 4), cBook)
        #IFDEF DEV
        shell("echo." & mid(l, len(sGame)+1, 4) & " b")
        #ELSE
        temp = cBook
        #ENDIF
        exit do
      end if
    loop
  end if
  '
  meilleurCoup = temp
  '
end function
